#pragma once
#include "ThreadJob.h"
#include "define.h"					// パケット作成のため

#define	CMD_BUFF_SIZE	(1024 * 64)	// ★ファイル送信コマンド/応答格納リングバッファのサイズ
#define TIMEOUT_FOR_RES	(2000)		// ★応答までのタイムアウト時間msec
// ★ファイル送信遷移状態
#define SEND_FILE_STAT_IDLE		0	// 初期状態
#define SEND_FILE_STAT_START	1	// ファイル送信開始要求応答待ち
#define SEND_FILE_STAT_ING		2	// ファイル送信中
#define SEND_FILE_STAT_END		3	// ファイル送信完了要求応答待ち
#define SEND_FILE_STAT_ABORT	4	// ファイル送信中断要求応答待ち

class CMySyncObject;				// CMySyncObjectを使うため
class CSendRecvThread;				// CSendRecvThreadを使うため
class CRingBuff;					// ★CRingBuffを使うため


class CSendFileThread :
	public CThreadJob
{
public:
	CSendFileThread(CSendRecvThread *pCSendRecvThread, LPSTR pszFileName);
	~CSendFileThread();

	CRingBuff	*m_pCRingBuffCmd;			// ★ファイル送信コマンド/応答格納リングバッファ

	UINT DoWork() override;
	BOOL IsZombie();						// このスレッドはゾンビ状態か
	BOOL SendFileAbort();					// ★ファイル送信の中断

private:
	LPSTR			m_pszFileName;			// 送信ファイル名
	CSendRecvThread	*m_pCSendRecvThread;	// 送受信スレッド(親スレッド)
	CMySyncObject	*m_pCMySyncObject;
	BOOL			m_fIamZombie;			// ゾンビ状態か(送信終了)

	int				m_iSendFileStatus;		// ★遷移状態 SEND_FILE_STAT_IDLE/SEND_FILE_STAT_START/SEND_FILE_STAT_ING/SEND_FILE_STAT_END
	char			m_cSendFileStartRes;	// ★ファイル送信開始要求応答(0:初期値 1:OK -1:エラー)
	char			m_cSendFileEndRes;		// ★ファイル送信完了要求応答(0:初期値 1:OK -1:エラー)
	char			m_cSendFileAbortRes;	// ★ファイル送信中断要求応答(0:初期値 1:OK -1:エラー)
	DWORD			m_dwStart;				// ★応答待ちタイムアウトのため要求を行った時刻
	int				m_fdSrcFile;			// ★送信ファイルディスクリプタ
	long			m_lFileLength;			// ★送信ファイルの長さ
	long			m_lTotalSentSize;		// ★送信済みバイト数
	int				m_iWaitCount;			// ★送信待ちをした回数

	int DoSendFileState();						// ★ファイル送信遷移状態に応じた処理を実行
	int SendFileStart();						// ★①ファイル送信開始処理
	int CheckSendFileStartRes(DWORD dwTimeOut);	// ★②ファイル送信開始要求応答チェック
	int CheckSendFileEndRes(DWORD dwTimeOut);	// ★⑤ファイル送信完了要求応答チェック
	int CheckSendFileAbortRes(DWORD dwTimeOut);	// ★ファイル送信中断要求応答チェック

	BOOL SendSendFileStartPacket(LPSTR pszFileName, int iFileLength);	// ①ファイル送信開始要求送信
	int SendFileIng();							// ★③④ファイル途中処理、ファイル送信完了処理
	BOOL SendSendFileAbortPacket();				// ★ファイル送信の中断要求送信

	int AnalyzeCmd();													// ★ファイル送信コマンド/応答パケットの解析
	int RecvSendFileStartResPacket(SendFileResRec *pbPacket);			// ★ファイル送信開始要求応答処理
	int RecvSendFileEndResPacket(SendFileResRec *pbPacket);				// ★ファイル送信完了要求応答処理
	int RecvSendFileAbortResPacket(SendFileResRec *pbPacket);			// ★ファイル送信中断要求応答処理
};

